
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Name:           UVunwrap.rb 
# Author:         Jan Sandstrom   www.pixero.com
# Date:           29 july 2011
# Description:	  Use to unwrap an object to project new UVs and transfer them back to the original object. 
#                 Works with ungrouped or inside grouped geometry.
# Based on Matt C's AutoUnfold script
# Copy paste UVs borrowed from Thomthom
# History:
#			Alpha 0.1.0 - 29 july 2011 (Proof of concept)
# To do:	Continue with making it work also for grouped geometry
#			Go through "AutoUnfold" code to find where it fails.
#			Implement another better unwrap technique. Like LSCM.		

require 'sketchup.rb'

#Sketchup.send_action "showRubyPanel:"

module UVunwrap

if not file_loaded?(__FILE__)
	submenu = UI.menu("Plugins").add_submenu("UV Unwrap")
	submenu.add_item("Unwrap") { Sketchup.active_model.select_tool UVunwrap::Unwrap.new() }  
	submenu.add_item("Show face pairs") { Sketchup.active_model.select_tool UVunwrap::Facepairs.new() }  
	submenu.add_item("Update UV's") { Sketchup.active_model.select_tool UVunwrap::Apply.new() }   
end
file_loaded(__FILE__)

########################################
class Unwrap 

	def initialize
	  @@ents=Sketchup.active_model.selection
	  @@bas_sel = []
	  @@state = 0
	  @@face1 = []
	  @@face2 = []
	  @@last_objId = 0
	end #initialize
	
	def get_faces(ents) #Get the faces of the selection
		faces = []
		definitions = []
		# Iterate selected entities
		ents.each { |e|
			if e.kind_of?(Sketchup::ComponentInstance) || e.kind_of?(Sketchup::Group)
				parent = e.definition if e.kind_of?(Sketchup::ComponentInstance)
				parent = e.entities.parent if e.kind_of?(Sketchup::Group)				
				next if definitions.include?(parent)					
				parent.entities.each { |e|
					if e.kind_of?(Sketchup::Face) 
						faces.push(e)
					end
				}
				definitions.push(parent)
			elsif e.kind_of?(Sketchup::Face) 
				faces.push(e)
			end
		}
		return faces 
	end
	
	def set_faceID(faces) #Set a Face ID to remember which face is which when reapplying the new UV's
		dict_value = 1
		faces.each { |f|									
			dict_name = "UV Unwrap"						
			dict_key = "Face ID"
			f.attribute_dictionary(dict_name, true)			
			f.set_attribute(dict_name,dict_key,dict_value)
			dict_value += 1
		}						
	end
	
	def set_objID(faces) #Set a Object ID to avoid mapping the wrong object when reapplying the new UV's	
		#model.get_attribute
		Sketchup.active_model.entities.each { |f|
		#faces.each { |f|
			compare_objId = f.get_attribute("UV Unwrap","Object ID") #get the highest objId in the scene.
			if compare_objId == nil
				compare_objId = 0
			end
			if compare_objId > @@last_objId
				@@last_objId = compare_objId
			end
		}
		dict_value = @@last_objId + 1
		faces.each { |f|									
			dict_name = "UV Unwrap"						
			dict_key = "Object ID"
			f.attribute_dictionary(dict_name, true)			
			f.set_attribute(dict_name,dict_key,dict_value)
		}						
	end

	def activate
		@@bas_sel = get_faces(@@ents)	
		set_objID(@@bas_sel) #Set an ObjID for every face	
		set_faceID(@@bas_sel) #Set an ID for every face		
		@@msg = @@bas_sel.length.to_s + " faces selected for unfolding. Select first face "
		Sketchup.set_status_text @@msg
		# Make a copy of the object to unwrap
		model = Sketchup.active_model  
		ent = model.active_entities  		
		group = ent.add_group(@@ents)  
		bounds = group.bounds 
		# Move the copy so it won't touch the original.
		v = [(bounds.width+100.mm), (bounds.height+100.mm), 0]
		tt = Geom::Transformation.new(v)		
		# defn = group.entities.parent
		# group2 = model.entities.add_instance(defn, group.transformation)
		group2 = group.copy
		group2.transform!(tt)
		# Put the copy on "UV layer" 
		layers = model.layers  		
		uv_layer = layers.add "UV layer"
		group2.layer = "UV layer"
		# Explode groups
		@@bas_sel = get_faces(group2.entities)		
		group2.entities.to_a.each{|e|
			e.explode if e.class==Sketchup::Group || e.class==Sketchup::ComponentInstance
		} 	
		#group.explode # How not to explode original group if any?
		group2.explode	
	end #activate
	
	def onMouseMove(flags, x, y, view)
		if @@face1.length == 0
				Sketchup.set_status_text "AutoUnfold:Select first face to unfold."
				return
		end
		if @@face1.length == 1
				Sketchup.set_status_text "Select second face."
				return
		end
	end # onMouseMove
	
	def onLButtonDown(flags, x, y, view)
		ph = Sketchup.active_model.active_view.pick_helper
		ph.do_pick x, y
		#ref = ph.picked_face
		ref = ph.leaf_at 0
		if ref.nil?
			Sketchup.active_model.selection.clear
			Sketchup.set_status_text "UV Unwrap:Select start face to unwrap."
			return
		end #ifnil
		
		if( ref.class==Sketchup::Face)
			if @@face1.length == 0
				@@state = 1
				@@face1 << ref
				Sketchup.active_model.selection.add ref
				Sketchup.set_status_text "Select second face"
				return
			end #if face1.length = 0
			if @@face1.length == 1
				@@state = 2
				@@face2 << ref
				Sketchup.set_status_text " Unwrap in progress"
				Sketchup.active_model.start_operation "Unwrap"
				st = uwrap(@@bas_sel,@@face1,@@face2)
			end #if face1.length = 1
			if st
				Sketchup.active_model.selection.add @@face1
			end # if st is not nil
		end # if face picked
		@@bas_sel = []
		@@face1 = []
		@@face2 = []
		
		Sketchup.active_model.commit_operation
	end # onLeftButtonDown
	
	def uwrap(bas_sel,face1,face2)
		#select all from basis selection that are coplanar with face1
		cop_faces = []	
		face1[0].all_connected.each do |f| 
			if f.class==Sketchup::Face
				if f.normal == face1[0].normal   
					cop_faces << f
				end #if faces coplanar
			end
		end #each bas_sel
		angle = face1[0].normal.angle_between face2[0].normal
		axis = Geom::intersect_plane_plane(face1[0].plane, face2[0].plane)
		#unfold first face to face2
		g = Sketchup.active_model.entities.add_group(cop_faces)
		g.transform! Geom::Transformation.rotation(axis[0], axis[1], angle)
		g.explode
		#set face2 as face1
		face1[0] = face2[0]

		cop_faces = []
		bas_sel.each do |f|
			if f.normal == face1[0].normal
				cop_faces << f
			end # if faces coplanar
		end #each bas_sel

		until cop_faces.length == bas_sel.length
			#find adjacent face to all coplanar faces
			cop_edges = []
			cop_faces.each do |f|
				f.edges.each do |e|
					cop_edges << e
				end # each cop_faces edges
			end # each cop_faces
			adj_faces = []
			cop_edges.each do |e|
				e.faces.each do |f|
					if bas_sel.include?(f)
						adj_faces << f
					end # if faces is in bas_sel
				end # each face
			end # each edge
			
			adj_faces = adj_faces - cop_faces
			#set first adj face as face2
			face2[0]=adj_faces[0]
			angle=[]
			axis=[]
			angle = face1[0].normal.angle_between face2[0].normal
			axis = Geom::intersect_plane_plane(face1[0].plane, face2[0].plane)
			#unfold first face to face2
			g = Sketchup.active_model.entities.add_group(cop_faces)
			g.transform! Geom::Transformation.rotation(axis[0], axis[1], angle)
			g.explode
			#set face2 as face1
			face1[0]=face2[0]
			
			#select all from basis selection that are coplanar with face1 
			cop_faces = []
			bas_sel.each do |f|
				if f.normal == face1[0].normal
					cop_faces<<f
				end
			end
			@@msg = "Unwrap in progress : " + cop_faces.length.to_s + " / " + bas_sel.length.to_s
			Sketchup.set_status_text @@msg
		end
		#unwrap all faces to plane xy
		angle=[]
		axis=[]
		angle = bas_sel[0].normal.angle_between [0,0,1]
		axis = Geom::intersect_plane_plane(bas_sel[0].plane, [0,0,1,0])
		g = Sketchup.active_model.entities.add_group(bas_sel)
		g.transform! Geom::Transformation.rotation(bas_sel[0].vertices[0].position, axis[1], angle)
		g.explode
		# Group the flattened faces and transform to z = 0
		uv_group = Sketchup.active_model.entities.add_group(cop_faces)
		uv_bboxmin = uv_group.bounds.min		
		v = [0, 0, -uv_bboxmin[2]]
		tt = Geom::Transformation.new(v)		
		uv_group.transform!(tt)
		
		
	end
end #class Unwrap
#################################3
class Apply

	def initialize		
	end #initialize
	
	def activate	
		ents = Sketchup.active_model.selection
		faces = []
		faces = get_faces(ents)		
		faces.each { |f|
			apply_UVs_to_face_pair(f)	
		}
	end#activate
	
	def get_faces(ents) #Get the faces of the selection
		faces = []
		definitions = []
		# Iterate selected entities
		ents.each { |e|
			if e.kind_of?(Sketchup::ComponentInstance) || e.kind_of?(Sketchup::Group)
				parent = e.definition if e.kind_of?(Sketchup::ComponentInstance)
				parent = e.entities.parent if e.kind_of?(Sketchup::Group)				
				next if definitions.include?(parent)					
				parent.entities.each { |e|
					if e.kind_of?(Sketchup::Face) 
						faces.push(e)
					end
				}
				definitions.push(parent)
			elsif e.kind_of?(Sketchup::Face) 
				faces.push(e)
			end
		}
		return faces 
	end
	
	def onMouseMove(flags, x, y, view)		
	end # onMouseMove
	
	def onLButtonDown(flags, x, y, view)
	end #onLeftButtonDown

	def get_faceID(face)	
		@@face_id = face.get_attribute("UV Unwrap","Face ID")
	end
	
	def get_objID(face)	
		@@obj_id = face.get_attribute("UV Unwrap", "Object ID")
	end
	
	def apply_UVs_to_face_pair(face)	
		face_id = get_faceID(face)
		obj_id = get_objID(face)	
		#Sketchup.active_model.entities.each{|e|
		get_faces(Sketchup.active_model.entities).each{|e|
				if e.get_attribute("UV Unwrap","Face ID") == face_id && e.get_attribute("UV Unwrap","Object ID") == obj_id
					#Apply UVs
					copy_uv(face)
					paste_uv(e)
				end
			} 
	end

	# Copy paste UVs borrowed from Thomthom
	# I've commented out the corners.length == 4 and self.quad_face? test,
	# since in this case we're using flattened geometry with planar projected UVs 
	# so there will be no distorted textures. 
	# Makes it work also for 3 sided geometry. :)
	
	@@clipboard = {}
	
	def normalizeUVQ(uvq)
		uvq.x /= uvq.z
		uvq.y /= uvq.z
		uvq.z = 1.0
		return uvq
	end
  
	def copy_uv(face) 
		return nil if face.nil?
		return nil if face.material.nil? || face.material.texture.nil?   
		corners = face_corners( face )
		#return nil unless corners.length == 4   
		tw = Sketchup.create_texture_writer
		uvh = face.get_UVHelper(true, true, tw)
    
		@@clipboard.clear
		@@clipboard[:uv] = []
		@@clipboard[:material] = face.material
		corners.each { |v|
		raw_uvq = uvh.get_front_UVQ( v.position )
		uvq = normalizeUVQ( raw_uvq )
		@@clipboard[:uv] << uvq
		}  
		true
	end
  
  # Finds the corners in a face, ignoring vertices between colinear edges.
	def face_corners(face)
		corners = []
		# We only check the outer loop, ignoring interior lines.
		face.outer_loop.edgeuses.each { |eu|
			# Ignore vertices that's between co-linear edges. When using .same_direction?
			unless eu.edge.line[1].parallel?(eu.next.edge.line[1])
				# Find which vertex is shared between the two edges.
				if eu.edge.start.used_by?(eu.next.edge)
					corners << eu.edge.start
				else
					corners << eu.edge.end
				end
			end
		}
		return corners
	end
	
	def paste_uv(face)
		return nil if @@clipboard.empty?
		return nil if @@clipboard[:material].deleted?  
		corners = face_corners( face )
		#return nil unless corners.length == 4     
		pts = []
		@@clipboard[:uv].each { |uvq|
			pt = corners.shift
			pts << pt
			pts << uvq
		}
		# Quick "hack" to be able to map faces with more than 4 vertices		
		pts2 = []
		pts2 = [pts[0], pts[1], pts[2], pts[3], pts[4], pts[5], pts[6], pts[7]]
		#face.position_material( @@clipboard[:material], pts, true )
		face.position_material( @@clipboard[:material], pts2, true )
    #end # for    
		true
	end
	
end #Apply

#################################3

#Use facepairs to visualize where a flattened face is on the original object. 
class Facepairs

	def initialize		
	end #initialize
	
	def activate		
	end#activate
	
	def onMouseMove(flags, x, y, view)		
	end # onMouseMove
	
	def onLButtonDown(flags, x, y, view)
		Sketchup.active_model.selection.clear
		ph = Sketchup.active_model.active_view.pick_helper
		ph.do_pick x, y
		ref = ph.leaf_at 0
		
		if( ref.class==Sketchup::Face)
			Sketchup.active_model.selection.add ref
			@@face_id = get_faceID(ref)
			@@obj_id = get_objID(ref)
			f2 = find_face_pair(ref)
		end
		Sketchup.active_model.commit_operation
	end #onLeftButtonDown
	
	def get_faceID(face)	
		@@face_id = face.get_attribute("UV Unwrap","Face ID")
	end
	
	def get_objID(face)	
		@@obj_id = face.get_attribute("UV Unwrap", "Object ID")
	end
	
	def find_face_pair(face)		
		Sketchup.active_model.selection.add( get_faces(Sketchup.active_model.entities).select{|e| 
			e.get_attribute("UV Unwrap","Face ID") == @@face_id && e.get_attribute("UV Unwrap","Object ID") == @@obj_id
			} 
		)
	end
	
	def get_faces(ents) #Get the faces of the selection
		faces = []
		definitions = []
		# Iterate selected entities
		ents.each { |e|
			if e.kind_of?(Sketchup::ComponentInstance) || e.kind_of?(Sketchup::Group)
				parent = e.definition if e.kind_of?(Sketchup::ComponentInstance)
				parent = e.entities.parent if e.kind_of?(Sketchup::Group)				
				next if definitions.include?(parent)					
				parent.entities.each { |e|
					if e.kind_of?(Sketchup::Face) 
						faces.push(e)
					end
				}
				definitions.push(parent)
			elsif e.kind_of?(Sketchup::Face) 
				faces.push(e)
			end
		}
		return faces 
	end	
		
end#class
##############################
end #module
